package org.processmining.models.operationalsupport.net.service;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.Writer;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketTimeoutException;
import java.net.UnknownHostException;
import java.util.SortedSet;
import java.util.TreeSet;

import javax.swing.SwingWorker;

import org.processmining.framework.plugin.events.Logger.MessageLevel;
import org.processmining.framework.util.socket.ServiceEnvironment;
import org.processmining.framework.util.socket.ServiceHandler;
import org.processmining.models.operationalsupport.Request;
import org.processmining.models.operationalsupport.Response;
import org.processmining.models.operationalsupport.net.OperationalSupportProtocol;
import org.processmining.models.operationalsupport.xml.OSXMLConverter;

/**
 * @author christian
 * 
 */
public abstract class OSServiceHandler implements ServiceHandler {

	private long REQUEST_COUNTER = 0;

	protected static String hostId;

	protected OSXMLConverter converter = new OSXMLConverter();

	private transient SortedSet<Integer> freePorts;

	private final int portOffset;

	private final int maxClients;

	static {
		try {
			InetAddress addr = InetAddress.getLocalHost();
			// Get IP Address
			byte[] ipAddr = addr.getAddress();
			// Get hostname
			String hostname = addr.getHostName();
			// Assemble host ID
			hostId = "";
			for (int i = 0; i < ipAddr.length; i++) {
				hostId += ipAddr[i];
				if ((i + 1) < ipAddr.length) {
					hostId += ".";
				}
			}
			hostId += " (" + hostname + ")";
		} catch (UnknownHostException e) {
		}
	}

	public OSServiceHandler(int portOffset, int maxClients) {
		this.portOffset = portOffset;
		this.maxClients = maxClients;
		setupPorts();
	}

	protected void setupPorts() {
		freePorts = new TreeSet<Integer>();
		for (int i = portOffset; i < portOffset + maxClients; i++) {
			freePorts.add(i);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.framework.remote.ServiceHandler#handleRequest(java.
	 * io.Reader, java.io.Writer)
	 */
	public void handleRequest(final ServiceEnvironment environment, BufferedReader in, PrintWriter out)
			throws IOException {
		// initiate session
		writeLine(out, OperationalSupportProtocol.HELO_SERVER_PROLOG);

		final Integer port;
		synchronized (freePorts) {
			while (freePorts.isEmpty()) {
				try {
					freePorts.wait();
				} catch (InterruptedException e) {
					// just ignore
				}
			}
			port = freePorts.first();
			freePorts.remove(port);
		}

		final ServerSocket serverSocket = new ServerSocket(port);

		writeLine(out, "OSService on host " + hostId);
		writeLine(out, Long.toString(REQUEST_COUNTER++)); // send the unique id for this request
		writeLine(out, Integer.toString(port)); // send the port to continue communication on
		writeLine(out, OperationalSupportProtocol.HELO_SERVER_EPILOG);

		// setup a new service to handle further communication asynchronously
		environment.log("Defering communication to port: " + port);

		// add the service to the list.

		SwingWorker<?, ?> worker = new SwingWorker<Object, Object>() {

			protected Object doInBackground() throws Exception {

				// Set the timeout to 15 minutes
				try {
					serverSocket.setSoTimeout(900000);
					Socket clientSocket = serverSocket.accept();
					BufferedReader in = new BufferedReader(new InputStreamReader(clientSocket.getInputStream()));
					PrintWriter out = new PrintWriter(clientSocket.getOutputStream(), true);
					handleServiceRequest(environment, in, out);
				} catch (SocketTimeoutException e) {
					environment.log("No Request received within 15 minutes after HELO", MessageLevel.ERROR);
				} catch (IOException e) {
					// abort and give up
					environment.log("Operational Support Server Failed: " + e.getMessage(), MessageLevel.ERROR);
					environment.log(e);
				} finally {
					try {
						serverSocket.close();
					} catch (IOException e) {
						environment.log("Operational Support Server Failed: " + e.getMessage(), MessageLevel.ERROR);
						environment.log(e);
					}
				}
				synchronized (freePorts) {
					freePorts.add(port);
					freePorts.notifyAll();
				}

				environment.log("Operational Support Server Finished", MessageLevel.DEBUG);
				return null;
			}

		};
		worker.run();

		environment.log("Finished handling the setup");

	}

	protected void writeLine(Writer out, String line) throws IOException {
		out.write(line + "\r\n");
		out.flush();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.framework.remote.ServiceHandler#handleRequest(java.
	 * io.Reader, java.io.Writer)
	 */
	private void handleServiceRequest(ServiceEnvironment environment, BufferedReader in, PrintWriter out)
			throws IOException {
		// some helpers and var initializations
		String received = null;
		String type = null;
		StringBuilder requestXml = new StringBuilder();
		String responseXml = null;

		// Session was initiated, just receive the request and send the response.

		while (((type = in.readLine()).equals(OperationalSupportProtocol.CLIENT_REQUEST_request_PROLOG) == false)
				&& ((type = in.readLine()).equals(OperationalSupportProtocol.CLIENT_REQUEST_INFO_PROLOG) == false)) {
			// wait for client request prolog
		}

		if (type.equals(OperationalSupportProtocol.CLIENT_REQUEST_request_PROLOG)) { // receive a request
			// receive request XML string
			while ((received = in.readLine()).equals(OperationalSupportProtocol.CLIENT_REQUEST_XML_EPILOG) == false) {
				requestXml.append(received);
				requestXml.append('\n'); // add proper newline terminator
			}
			// signal receipt
			writeLine(out, OperationalSupportProtocol.SERVER_REQUEST_XML_RECEIVED);
			// process request
			Request request;
			Response response;
			// unmarshal request from XML
			try {
				request = converter.requestFromXML(requestXml.toString());
			} catch (Exception e) {
				sendError(out, "Exception thrown when unmarshalling your request:", e);
				return; // give up
			}
			// process request
			try {
				response = getResult(environment, request);
			} catch (Exception e) {
				sendError(out, "Exception thrown processing your request:", e);
				return; // give up
			}
			// marshal result to XML
			try {
				responseXml = converter.responseToXML(response);
			} catch (Exception e) {
				sendError(out, "Exception thrown when marshalling the result to your request:", e);
				return; // give up
			}
		}

		if (type.equals(OperationalSupportProtocol.CLIENT_REQUEST_INFO_PROLOG)) { // receive a request
			// receive request XML string
			String provider = "";
			StringBuilder info = new StringBuilder();
			writeLine(out, OperationalSupportProtocol.SERVER_INFO_PROVIDER);
			provider = in.readLine(); // get the provider id
			while ((received = in.readLine()).equals(OperationalSupportProtocol.CLIENT_REQUEST_XML_EPILOG) == false) {
				info.append(received);
				info.append('\n'); // add proper newline terminator
			}
			responseXml = processInfo(environment, provider, info.toString());
			// signal receipt
			writeLine(out, OperationalSupportProtocol.SERVER_REQUEST_XML_RECEIVED);
			// process info

		}

		// signal success
		writeLine(out, OperationalSupportProtocol.SERVER_RESPONSE_POSITIVE);
		// send result (sandwiched in prolog & epilog)
		writeLine(out, OperationalSupportProtocol.SERVER_RESULT_PROLOG);
		if (responseXml != null) {
			writeLine(out, responseXml);
		}
		writeLine(out, OperationalSupportProtocol.SERVER_RESULT_EPILOG);
		// wait for goodbye (i.e. all right received)
		while ((received = in.readLine()).equals(OperationalSupportProtocol.CLIENT_GOODBYE) == false) {
			// wait for client goodbye
		}
		// send own goodbye and terminate
		writeLine(out, OperationalSupportProtocol.SERVER_GOODBYE);
		// done.
	}

	protected void sendError(PrintWriter out, String reason, Exception e) throws IOException {
		writeLine(out, OperationalSupportProtocol.SERVER_RESPONSE_NEGATIVE);
		writeLine(out, reason);
		writeLine(out, e.getMessage());
		writeLine(out, e.toString());
		e.printStackTrace(out);
		writeLine(out, " ...giving up and closing connection!");
		writeLine(out, OperationalSupportProtocol.SERVER_GOODBYE);
	}

	protected abstract Response getResult(ServiceEnvironment environment, Request request);

	protected abstract String processInfo(ServiceEnvironment environment, String provider, String info);

}
